from colorama import Fore, Style

from oarepo_cli.model.install.steps.alembic import CreateAlembicModelStep
from oarepo_cli.model.install.steps.install import InstallModelStep
from oarepo_cli.model.install.steps.test_model import RunTestsModelStep
from oarepo_cli.model.install.steps.update_index import UpdateSearchIndexModelStep
from oarepo_cli.wizard import RadioStep, Wizard
from oarepo_cli.wizard.docker import DockerRunner


class InstallModelWizard(Wizard):
    def __init__(self, runner: DockerRunner, *, model_support, site_support):
        self.model_support = model_support
        self.site_support = site_support

        super().__init__(
            RadioStep(
                "run_tests",
                options={
                    "run": f"{Fore.GREEN}Run tests{Style.RESET_ALL}",
                    "skip": f"{Fore.RED}Skip tests{Style.RESET_ALL}",
                },
                default="run",
                heading=f"""
            Before installing the model, it is wise to run the test to check that the model is ok.
            If the tests fail, please fix the errors and run this command again.
                """,
                force_run=True,
            ),
            *runner.wrap_docker_steps(
                RunTestsModelStep(),
                InstallModelStep(),
                CreateAlembicModelStep(),
                UpdateSearchIndexModelStep(),
            ),
        )



import click as click

from oarepo_cli.utils import commit_git, with_config

from ...site.site_support import SiteSupport
from ...wizard.docker import DockerRunner
from .wizard import AddLocalWizard


@click.command(
    name="add",
    help="""Add a local package:
    <name>               ... pypi name of the package
    """,
)
@click.option("--site-name", help="Site where to install the package")
@click.argument("name")
@click.argument("github_url", required=False)
@click.option("--branch", help="Branch")
@with_config(config_section=lambda name, **kwargs: ["local", name])
def add_local(
    cfg=None,
    step=None,
    no_input=False,
    silent=False,
    verbose=False,
    steps=False,
    site_name=None,
    github_url=None,
    branch=None,
    **kwargs,
):
    commit_git(
        cfg.project_dir,
        f"before-local-clone-{cfg.section}",
        f"Committed automatically before package {cfg.section} has been cloned",
    )
    cfg["local_dir"] = f"local/{cfg.section_path[-1]}"
    cfg["github_clone_url"] = github_url
    cfg["branch"] = branch

    site_support = SiteSupport(cfg, site_name)

    local_sites = cfg.setdefault("sites", [])
    if site_name not in local_sites:
        local_sites.append(site_support.site_name)

    runner = DockerRunner(cfg, no_input)
    wizard = AddLocalWizard(runner, site_support)
    if steps:
        wizard.list_steps()
        return

    wizard.run_wizard(
        cfg, selected_steps=step, no_input=no_input, silent=silent, verbose=verbose
    )
    commit_git(
        cfg.project_dir,
        f"after-local-clone-{cfg.section}",
        f"Committed automatically after package {cfg.section} has been cloned",
    )



import shutil

from oarepo_cli.site.mixins import SiteWizardStepMixin
from oarepo_cli.wizard import WizardStep


class CheckSiteStep(SiteWizardStepMixin, WizardStep):
    def __init__(self, clean=False):
        super().__init__()
        self.clean = clean

    def after_run(self):
        # site will be newly installed, so need to get rid of compiled stuff
        # doing it before site installation as it will place 'invenio.cfg'
        # there
        if self.site_support.invenio_instance_path.exists():
            shutil.rmtree(self.site_support.invenio_instance_path)

        self.site_support.install_site()

    def should_run(self):
        return self.clean or not self.site_support.site_ok()



import json
import re
from pathlib import Path

from oarepo_cli.utils import copy_tree

# Taken from Invenio-cli
#
# this and the following were taken from:
# https://github.com/inveniosoftware/invenio-cli/blob/0a49d438dc3c5ace872ce27f8555b401c5afc6e7/invenio_cli/commands/local.py#L45
# and must be called from the site directory
#
# The reason is that symlinking functionality is only part of invenio-cli
# and that is dependent on pipenv, which can not be used inside alpine
# (because we want to keep the image as small as possible, we do not install gcc
# and can only use compiled native python packages - like cairocffi or uwsgi). The
# version of these provided in alpine is slightly lower than the one created by Pipenv
# - that's why we use plain invenio & pip here.
#
# Another reason is that invenio-cli is inherently unstable when non-rdm version
# is used - it gets broken with each release.


def register_less_components(site, invenio_instance_path):
    site.call_invenio(
        "oarepo",
        "assets",
        "less-components",
        f"{invenio_instance_path}/less-components.json",
    )
    data = json.loads(Path(f"{invenio_instance_path}/less-components.json").read_text())
    components = list(set(data["components"]))
    theme_config_file = site.site_dir / "assets" / "less" / "theme.config"
    theme_data = theme_config_file.read_text()
    for c in components:
        match = re.search("^@" + c, theme_data, re.MULTILINE)
        if not match:
            match = theme_data.index("/* @my_custom_component : 'default'; */")
            theme_data = (
                theme_data[:match] + f"\n@{c}: 'default';\n" + theme_data[match:]
            )
    theme_config_file.write_text(theme_data)


def load_watched_paths(paths_json, extra_paths):
    watched_paths = {}
    with open(paths_json) as f:
        for target, paths in json.load(f).items():
            if target.startswith("@"):
                continue
            for pth in paths:
                watched_paths[pth] = target
    for e in extra_paths:
        source, target = e.split("=", maxsplit=1)
        watched_paths[source] = target
    return watched_paths


def copy_watched_paths(watched_paths, destination):
    destination.mkdir(parents=True, exist_ok=True)
    for source, target in watched_paths.items():
        copy_tree(Path(source).absolute(), destination.absolute() / target)



import os

from oarepo_cli.upgrade.steps.upgrade_nrp import upgrade_venv
from oarepo_cli.wizard import WizardStep


class UpgradeDockerNRPStep(WizardStep):
    def should_run(self):
        return self.data.running_in_docker

    def after_run(self):
        upgrade_venv(os.environ["NRP_VENV"])



import click as click

from oarepo_cli.run.wizard import RunSiteWizard
from oarepo_cli.site.site_support import SiteSupport
from oarepo_cli.utils import with_config
from oarepo_cli.wizard.docker import DockerRunner


@click.command(name="run", help="Run the server")
@click.option("-c", "--celery")
@click.option("--site", default=None, required=False)
@with_config()
def run_server_command(
    cfg=None,
    site=None,
    no_input=False,
    silent=False,
    step=None,
    steps=False,
    verbose=False,
    **kwargs
):
    site_support = SiteSupport(cfg, site)

    model_sites = cfg.setdefault("sites", [])
    if site_support.site_name not in model_sites:
        model_sites.append(site_support.site_name)
    cfg.save()

    runner = DockerRunner(cfg, no_input)
    wizard = RunSiteWizard(runner, site_support=site_support)

    if steps:
        wizard.list_steps()
        return

    wizard.run_wizard(
        cfg, no_input=no_input, silent=silent, selected_steps=step, verbose=verbose
    )



import re
import subprocess

from oarepo_cli.site.mixins import SiteWizardStepMixin
from oarepo_cli.wizard import RadioStep, WizardStep


class InitDatabaseStep(SiteWizardStepMixin, WizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            RadioStep(
                "init_database",
                options={"yes": "Yes", "no": "No"},
                default="yes",
            ),
            heading="""
If the database has not been initialized, I can do it now - 
this will delete all the previous data in the database.

Should I do it?
            """,
            **kwargs,
        )

    def after_run(self):
        if self.data["init_database"] == "yes":
            self.site_support.call_invenio("db", "create")
        self.check_db_initialized(raise_error=True)

    def check_db_initialized(self, raise_error=False):
        try:
            output = self.site_support.call_invenio(
                "alembic",
                "current",
                grab_stdout=True,
                raise_exception=True,
            )
        except subprocess.CalledProcessError:
            raise Exception(
                "Alembic initialization failed. This could mean that the database "
                "does not exist or that there is already an incompatible alembic "
                "(previous repository) present in the database. Please fix the problem "
                "and try again"
            )
        if re.search("[a-zA-Z0-9]{12,24}\s+->\s+[a-zA-Z0-9]{12,24}", output):
            return True
        if raise_error:
            raise Exception(
                f"DB initialization failed. Expected alembic heads, got\n{output}"
            )
        return False

    def should_run(self):
        return not self.check_db_initialized(raise_error=False)



from oarepo_cli.develop.steps.check_dependencies import CheckDependenciesStep
from oarepo_cli.develop.steps.check_site import CheckSiteStep
from oarepo_cli.develop.steps.check_ui import CheckUIStep
from oarepo_cli.develop.steps.check_venv import CheckVirtualenvStep
from oarepo_cli.site.site_support import SiteSupport
from oarepo_cli.wizard import Wizard
from oarepo_cli.wizard.docker import DockerRunner


class BuildWizard(Wizard):
    def __init__(
        self, runner: DockerRunner, site_support: SiteSupport, production: bool
    ):
        self.site_support = site_support
        super().__init__(
            *runner.wrap_docker_steps(
                CheckVirtualenvStep(clean=True),
                CheckDependenciesStep(clean=False, require_up_to_date=production),
                CheckSiteStep(clean=True),
                CheckUIStep(production=production),
            ),
        )



from oarepo_cli.site.mixins import SiteWizardStepMixin
from oarepo_cli.wizard import WizardStep


class NextStepsStep(SiteWizardStepMixin, WizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            heading=lambda data: f"""
The repository skeleton has been created and dependencies installed.
To check that everything has been installed successfully you may want
to start your new repository by running

    nrp run
    
and point your browser to 

    https://localhost:5000/
    
After doing so, add your model by calling

    nrp model add <modelname>
    
edit the contents of metadata.yaml to suit your needs and run

    nrp model compile <modelname>
    nrp model install <modelname>
    
If you run invenio again and head to https://localhost:5000/api/<modelname>
an empty listing should be returned.

In the next step, add a UI application by calling

    nrp ui add <uiname>
    nrp ui compile <uiname>
    nrp ui install <uiname>

After restarting the server, a UI will be available at 
https://localhost:5000/<uiname>
            """,
            **kwargs,
        )

    def should_run(self):
        return True



import click

from oarepo_cli.initialize.wizard import InitializeWizard
from oarepo_cli.utils import with_config


@click.command(
    name="initialize",
    help="""
Initialize the whole repository structure. Required arguments:
    <project_dir>   ... path to the output directory
""",
)
@click.option(
    "--no-site",
    default=False,
    is_flag=True,
    type=bool,
    help="Do not create default site",
)
@click.option("--python", required=False)
@with_config(project_dir_as_argument=True)
def initialize_command(
    *,
    context=None,
    cfg=None,
    no_site=False,
    python=None,
    step=None,
    steps=False,
    no_input=False,
    silent=False,
    verbose=False,
    **kwargs
):
    cfg["python"] = python or "python3.9"
    initialize_wizard = InitializeWizard()
    if steps:
        initialize_wizard.list_steps()
        return

    initialize_wizard.run_wizard(
        cfg, selected_steps=step, no_input=no_input, silent=silent, verbose=verbose
    )



import json
import shutil

from oarepo_cli.utils import run_cmdline
from oarepo_cli.wizard import WizardStep


class UpgradeNRPStep(WizardStep):
    def should_run(self):
        return True

    def after_run(self):
        for venv_dir in (self.data.project_dir / ".venv").glob("*"):
            if not venv_dir.is_dir():
                continue
            if venv_dir.name == "oarepo-model-builder":
                shutil.rmtree(venv_dir)
                continue
            if not (venv_dir / "bin" / "python").exists():
                continue
            upgrade_venv(venv_dir)


def upgrade_venv(venv_dir):
    # run
    packages = run_cmdline(
        "./bin/pip",
        "list",
        "--outdated",
        "--format",
        "json",
        cwd=venv_dir,
        grab_stdout=True,
        grab_stderr=False,
        raise_exception=True,
    )
    packages = json.loads(packages)
    obsolete_packages = [
        f"{p['name']}=={p['latest_version']}"
        for p in packages
        if p["name"].startswith("oarepo") or p["name"].startswith("nrp")
    ]
    if obsolete_packages:
        run_cmdline(
            "./bin/pip",
            "install",
            "-U",
            *obsolete_packages,
            cwd=venv_dir,
            raise_exception=True,
        )



import click

from .add.cli import add_site
from .list import list_sites


@click.group(help="Repository site related tools", name="site")
def site_commands():
    pass


site_commands.add_command(add_site)
site_commands.add_command(list_sites)



import sys

import yaml

from oarepo_cli.config import MonorepoConfig
from oarepo_cli.utils import run_nrp_in_docker, run_nrp_in_docker_compose
from oarepo_cli.wizard import WizardStep


class RunInContainerStep(WizardStep):
    def __init__(self, steps, in_compose=True, interactive=False, site=None, name=None):
        self.steps = steps or []
        self.in_compose = in_compose
        self.interactive = interactive
        self.site = site

    def should_run(self):
        return True

    @property
    def site_support(self):
        from oarepo_cli.site.site_support import SiteSupport

        if self.site:
            return SiteSupport(self.data, self.site)
        return self.root.site_support

    def run(self, selected_steps=None):
        cmd = sys.argv[1:]
        # remove project dir as it is added by docker itself
        for idx, c in enumerate(cmd):
            if c == "--project-dir":
                cmd.pop(idx)
                cmd.pop(idx)
                break
        for step in self.steps:
            if selected_steps and step not in selected_steps:
                continue
            cmd.append("--step")
            cmd.append(step.name)
        if self.site:
            cmd.append("--site")
            cmd.append(self.site)
        if self.in_compose:
            run_nrp_in_docker_compose(
                self.site_support.site_dir, *cmd, interactive=self.interactive
            )
        else:
            run_nrp_in_docker(self.data.project_dir, *cmd, interactive=self.interactive)
        # the config could have been changed during the run, so reload it
        self.data.load()

    @property
    def name(self):
        return "_".join(x.name for x in self.steps)


class DockerRunner:
    def __init__(self, cfg: MonorepoConfig, no_input):
        self.cfg = cfg
        self.no_input = no_input

        # load user config
        user_config_path = cfg.path.parent / ".oarepo-user.yaml"
        if user_config_path.exists():
            with open(user_config_path, "r") as f:
                user_config = yaml.safe_load(f)
        else:
            user_config = {}

        # if not overriden by user, load from user config file
        if self.use_docker is None:
            cfg.use_docker = user_config.get("use_docker", None)

        # if user has not yet selected if running through docker, force
        if self.use_docker is None and not self.running_in_docker and not self.no_input:
            cfg.use_docker = (
                input(
                    "I can run all the steps inside a container. Should I do so? [y/n]"
                )
                == "y"
            )
            user_config["use_docker"] = cfg.use_docker
            # write
            with open(user_config_path, "w") as f:
                yaml.safe_dump(user_config, f)

    @property
    def running_in_docker(self):
        return self.cfg.running_in_docker

    @property
    def use_docker(self):
        return self.cfg.use_docker

    def wrap_docker_steps(self, *steps, in_compose=True, interactive=False, site=None):
        if not self.use_docker or self.running_in_docker:
            return steps
        return [
            RunInContainerStep(
                steps, in_compose=in_compose, interactive=interactive, site=site
            )
        ]



import json
import os
import shutil
import subprocess
import tempfile
from pathlib import Path

import requirements
import tomli_w
from dotenv import dotenv_values
from minio import Minio

from oarepo_cli.config import MonorepoConfig
from oarepo_cli.package_versions import OAREPO_VERSION, PYTHON_VERSION
from oarepo_cli.utils import run_cmdline


class SiteSupport:
    def __init__(self, config: MonorepoConfig, site_section=None):
        if config.section_path[0] == "sites":
            self.site = config
            self.site_name = config.section_path[-1]
            self.config = config
            return
        elif not site_section:
            sites = config.whole_data.get("sites", {})
            if len(sites) == 1:
                site_section = next(iter(sites.keys()))
            else:
                raise RuntimeError("no or more sites, please specify --site or similar")

        self.site = config.whole_data.get("sites", {})[site_section]
        self.site_name = site_section
        self.config = config.clone(["sites", self.site_name])

    @property
    def site_dir(self):
        return Path(self.config.project_dir) / self.site["site_dir"]

    @property
    def python(self):
        if self.config.running_in_docker:
            python = "python3"
        else:
            python = self.config.whole_data["config"]["python"]
        #
        # python_base_package = run_cmdline(
        #     python,
        #     "-c",
        #     "import sys; print(sys.base_prefix)",
        #     grab_stdout=True,
        # )
        # return python_base_package.strip() + "/bin/" + python
        return python

    def call_pdm(self, *args, **kwargs):
        pdm_binary = self.site.get("pdm_binary", "pdm")
        return run_cmdline(
            pdm_binary,
            *args,
            cwd=self.site_dir,
            environ={"PDM_IGNORE_ACTIVE_VENV": "1"},
            **kwargs,
        )

    @property
    def virtualenv(self):
        return Path(os.environ.get("INVENIO_VENV", self.site_dir / ".venv"))

    @property
    def invenio_instance_path(self):
        return Path(
            os.environ.get(
                "INVENIO_INSTANCE_PATH", self.virtualenv / "var" / "instance"
            )
        )

    def call_pip(self, *args, **kwargs):
        return run_cmdline(
            self.virtualenv / "bin" / "pip",
            *args,
            **{
                "cwd": self.site_dir,
                "raise_exception": True,
                **kwargs,
            },
        )

    def call_invenio(self, *args, **kwargs):
        return run_cmdline(
            self.virtualenv / "bin" / "invenio",
            *args,
            **{
                "cwd": self.site_dir,
                "raise_exception": True,
                **kwargs,
            },
        )

    def get_site_local_packages(self):
        models = [
            model_name
            for model_name, model_section in self.config.whole_data.get(
                "models", {}
            ).items()
            if self.config.section in model_section.get("sites", [])
        ]
        uis = [
            ui_name
            for ui_name, ui_section in self.config.whole_data.get("ui", {}).items()
            if self.config.section in ui_section.get("sites", [])
        ]
        local_packages = [
            local_name
            for local_name, local_section in self.config.whole_data.get(
                "local", {}
            ).items()
            if self.config.section in local_section.get("sites", [])
        ]
        return models, uis, local_packages

    def venv_ok(self):
        if self.virtualenv.exists():
            # check if the virtualenv is usable - try to run python
            try:
                run_cmdline(
                    self.virtualenv / "bin" / "python",
                    "--version",
                    raise_exception=True,
                )
                run_cmdline(
                    self.virtualenv / "bin" / "pip",
                    "list",
                    raise_exception=True,
                    grab_stdout=True,
                )
                return True
            except:
                pass
        return False

    def check_and_create_virtualenv(self, clean=False):
        if not self.venv_ok():
            clean = True

        if clean and self.virtualenv.exists():
            shutil.rmtree(self.virtualenv)

        cmdline = [
            self.python,
            "-m",
            "venv",
        ]
        if self.config.running_in_docker:
            # alpine image has a pre-installed deps, keep them here
            cmdline.append("--system-site-packages")

        run_cmdline(
            *cmdline, str(self.virtualenv), cwd=self.site_dir, raise_exception=True
        )
        self.call_pip(
            "install",
            "-U",
            "--no-input",
            "setuptools",
            "pip",
            "wheel",
        )

    def _get_oarepo_dependencies(self, oarepo, system_packages):
        self.call_pip("download", "--no-deps", "--no-binary=:all:", oarepo, cwd="/tmp")
        tar_name = "/tmp/" + oarepo.replace("==", "-") + ".tar.gz"
        # extract the tar
        with tempfile.TemporaryDirectory() as temp_dir:
            import tarfile

            tf = tarfile.open(tar_name, mode="r:gz")
            tf.extractall(path=temp_dir)
            content_dir = temp_dir + "/" + oarepo.replace("==", "-")
            run_cmdline(
                self.virtualenv / "bin" / "python",
                "setup.py",
                "egg_info",
                cwd=content_dir,
            )
            requires = (
                Path(content_dir) / "oarepo.egg-info" / "requires.txt"
            ).read_text()
            requires = requires.split("\n\n")[0].split("\n")
            filtered_requires = []
            for r in requires:
                if r.split("==")[0].lower() not in system_packages:
                    filtered_requires.append(r)
            return filtered_requires

    def _install_oarepo_dependencies(self, oarepo, system_packages):
        print("Install oarepo dependencies called")
        requirements_to_install = self._get_oarepo_dependencies(oarepo, system_packages)
        print("Requirements to install", requirements_to_install)

        # this is needed to fix installation problems on osx (not all requirements
        # seems to be built inside oarepo package for darwin architecture)
        self.call_pip("install", "--force-reinstall", "ipython")

        # and call the following with force reinstall as ipython might override something
        with tempfile.NamedTemporaryFile(
            mode="wt", suffix="-requirements.txt"
        ) as temp_file:
            temp_file.write("\n".join(requirements_to_install))
            temp_file.flush()
            self.call_pip(
                "install", "--no-deps", "--force-reinstall", "-r", temp_file.name
            )
        #
        # install oarepo, but do not install dependencies as these have been already installed
        #
        self.call_pip("install", "--no-deps", "--force-reinstall", oarepo)

        # hack: add an empty version of uritemplate.py,
        # needs to be removed when invenio-oauthclient gets updated
        self.call_pip(
            "install",
            "--force-reinstall",
            str(Path(__file__).parent / "uritemplate.py-1.999.999.tar.gz"),
        )

    def site_ok(self):
        try:
            self.call_invenio("--help", raise_exception=True, grab_stdout=True)
        except:
            return False

        models, uis, local_packages = self.get_site_local_packages()
        invenio_ts = (self.virtualenv / "bin" / "invenio").lstat().st_mtime
        print(f"Invenio has timestamp {invenio_ts}")
        if self.packages_newer(models, "models", invenio_ts):
            return False
        if self.packages_newer(uis, "ui", invenio_ts):
            return False
        if self.packages_newer(local_packages, "local", invenio_ts):
            return False
        return True

    def packages_newer(self, packages, package_folder, timestamp):
        for package in packages:
            package_dir = (
                self.site_dir.absolute().parent.parent / package_folder / package
            )
            if not package_dir.exists():
                continue
            for fn in package_dir.glob("*"):
                if "egg" in fn.name:
                    continue
                if fn.lstat().st_mtime > timestamp:
                    print(
                        f"Package {fn} with timestamp {fn.lstat().st_mtime} is newer than {timestamp}"
                    )
                    return True
        return False

    def install_site(self):
        identified_requirements = (
            (self.site_dir / "requirements.txt").read_text().splitlines()
        )
        oarepo = identified_requirements[0]
        no_oarepo = identified_requirements[1:]

        system_packages = self._get_system_packages()
        self._install_extra_packages_if_not_installed()
        self._install_oarepo_dependencies(oarepo, system_packages)
        self._install_fake_oarepo(oarepo)
        self._install_requirements(no_oarepo)

        instance_dir = self.invenio_instance_path
        if not instance_dir.exists():
            instance_dir.mkdir(parents=True)
        if not (instance_dir / "invenio.cfg").exists():
            (instance_dir / "invenio.cfg").symlink_to(self.site_dir / "invenio.cfg")
        if not (instance_dir / "variables").exists():
            (instance_dir / "variables").symlink_to(self.site_dir / "variables")

        # now install all the local packages without dependencies as these were already
        # collected in the requirements.txt

        # main site
        site_package_dir = self.site_dir.absolute() / "site"
        for f in site_package_dir.glob("*.egg-info"):
            shutil.rmtree(f)
        self.call_pip(
            "install", "-U", "--no-input", "--no-deps", "-e", str(site_package_dir)
        )

        # models and uis
        models, uis, local_packages = self.get_site_local_packages()
        self._install_package(models, "models")
        self._install_package(uis, "ui")
        self._install_package(local_packages, "local")

        # touch invenio to mark the installation timestamp
        (self.virtualenv / "bin" / "invenio").touch()

    def _get_system_packages(self):
        tmpd = tempfile.mkdtemp()
        try:
            cmdline = [
                self.python,
                "-m",
                "venv",
            ]
            if self.config.running_in_docker:
                # alpine image has a pre-installed deps, keep them here
                cmdline.append("--system-site-packages")

            run_cmdline(*cmdline, tmpd, raise_exception=True, cwd=tmpd)
            run_cmdline(
                Path(tmpd) / "bin" / "pip",
                "install",
                "-U",
                "--no-input",
                "setuptools",
                "pip",
                "wheel",
                cwd=tmpd,
                no_environment=True,
            )
            initial_json = json.loads(
                run_cmdline(
                    Path(tmpd) / "bin" / "pip",
                    "list",
                    "--format",
                    "json",
                    grab_stdout=True,
                    cwd=tmpd,
                    no_environment=True,
                )
            )
            return [x["name"].lower() for x in initial_json]
        finally:
            shutil.rmtree(tmpd)

    def _install_fake_oarepo(self, oarepo):
        # create a shadow oarepo packages without any dependencies
        # to prevent dependency clash
        oarepo_version = oarepo.split("==", maxsplit=1)[1]
        tmpd = tempfile.mkdtemp()
        try:
            (Path(tmpd) / "setup.py").write_text(
                f"""
from setuptools import setup

setup(name='oarepo',
      version='{oarepo_version}',
      description='OARepo dependencies',
      url='http://github.com/oarepo/oarepo',
      author='CESNET z.s.p.o',
      author_email='miroslav.simek@cesnet.cz',
      license='MIT',
      packages=[],
      zip_safe=True)            
            """
            )
            self.call_pip("install", tmpd)
        finally:
            shutil.rmtree(tmpd)

    def _install_extra_packages_if_not_installed(self):
        # extra packages - modify oarepo not to exclude them
        extra_packages = [
            "libcst",
            "cchardet",
            "uwsgi",
            "ruamel.yaml.clib",
            "cairocffi",
            "cffi",
            "packaging",
            "pyparsing",
        ]
        self.call_pip("install", *extra_packages)

    def _install_requirements(self, package_list):
        with tempfile.NamedTemporaryFile(suffix="-requirements.txt", mode="w") as f:
            f.write("\n".join(package_list))
            f.flush()
            self.call_pip(
                "install",
                "-U",
                "--no-input",
                "--no-deps",
                "-r",
                f.name,
            )

    def _install_package(self, packages, package_folder):
        for package in packages:
            package_dir = (
                self.site_dir.absolute().parent.parent / package_folder / package
            )
            if not package_dir.exists():
                continue
            for f in package_dir.glob("*.egg-info"):
                shutil.rmtree(f)
            self.call_pip(
                "install", "-U", "--no-input", "--no-deps", "-e", str(package_dir)
            )

    def build_ui(self, production=False):
        from oarepo_cli.site.assets import (
            copy_watched_paths,
            load_watched_paths,
            register_less_components,
        )

        invenio_instance_path = self.invenio_instance_path.resolve()

        shutil.rmtree(invenio_instance_path / "assets", ignore_errors=True)
        shutil.rmtree(invenio_instance_path / "static", ignore_errors=True)

        Path(invenio_instance_path / "assets").mkdir(parents=True)
        Path(invenio_instance_path / "static").mkdir(parents=True)

        register_less_components(self, invenio_instance_path)

        self.call_invenio(
            "oarepo",
            "assets",
            "collect",
            f"{invenio_instance_path}/watch.list.json",
        )
        self.call_invenio(
            "webpack",
            "clean",
            "create",
        )
        self.call_invenio(
            "webpack",
            "install",
        )

        assets = (self.site_dir / "assets").resolve()
        static = (self.site_dir / "static").resolve()

        watched_paths = load_watched_paths(
            invenio_instance_path / "watch.list.json",
            [f"{assets}=assets", f"{static}=static"],
        )

        copy_watched_paths(watched_paths, invenio_instance_path)

        self.call_invenio("webpack", "build", *(["--production"] if production else []))

        # do not allow Clean plugin to remove files
        webpack_config = (
            invenio_instance_path / "assets" / "build" / "webpack.config.js"
        ).read_text()
        webpack_config = webpack_config.replace("dry: false", "dry: true")
        (invenio_instance_path / "assets" / "build" / "webpack.config.js").write_text(
            webpack_config
        )

    def ui_ok(self):
        # check that there is a manifest.json there
        manifest = self.invenio_instance_path / "static" / "dist" / "manifest.json"
        if not manifest.exists():
            return False
        try:
            json_data = json.loads(manifest.read_text())
            if json_data.get("status") != "done":
                return False
        except:
            return False
        return True

    def rebuild_site(self, clean=False, build_ui=False):
        self.check_and_create_virtualenv(clean=clean)
        self.build_dependencies()
        self.install_site()
        if build_ui:
            self.build_ui()

    def build_dependencies(self):
        # create pyproject.toml file
        models, uis, local_packages = self.get_site_local_packages()
        extras = [
            *[
                f"{model} @ file:///${{PROJECT_ROOT}}/../../models/{model}"
                for model in models
                if (self.site_dir.parent.parent / "models" / model).exists()
            ],
            *[
                f"{ui} @ file:///${{PROJECT_ROOT}}/../../ui/{ui}"
                for ui in uis
                if (self.site_dir.parent.parent / "ui" / ui).exists()
            ],
            *[
                f"{local} @ file:///${{PROJECT_ROOT}}/../../local/{local}"
                for local in local_packages
                if (self.site_dir.parent.parent / "local" / local).exists()
            ],
            "site @ file:///${PROJECT_ROOT}/site",
        ]
        # generate requirements just for oarepo package
        oarepo_requirements = self.generate_requirements([])
        oarepo_requirements = list(requirements.parse(oarepo_requirements))

        # get the current version of oarepo
        oarepo_requirement = [x for x in oarepo_requirements if x.name == "oarepo"][0]

        # generate requirements for the local packages as well
        all_requirements = self.generate_requirements(extras)
        all_requirements = list(requirements.parse(all_requirements))

        # now make the difference of those two (we do not want to have oarepo dependencies in the result)
        # as oarepo will be installed to virtualenv separately (to handle system packages)
        oarepo_requirements_names = {x.name for x in oarepo_requirements}
        non_oarepo_requirements = [
            x for x in all_requirements if x.name not in oarepo_requirements_names
        ]

        # remove local packages
        non_oarepo_requirements = [
            x for x in non_oarepo_requirements if "file://" not in x.line
        ]

        # and generate final requirements
        resolved_requirements = "\n".join(
            [oarepo_requirement.line, *[x.line for x in non_oarepo_requirements]]
        )
        (self.site_dir / "requirements.txt").write_text(resolved_requirements)

    def require_dependencies_up_to_date(self):
        reqs_file = self.site_dir / "requirements.txt"
        if not reqs_file.exists():
            raise RuntimeError(
                "No requirements.txt has been found! You need to run nrp build first"
            )
        reqs_file_backup = self.site_dir / "requirements-previous.txt"
        reqs_file_backup.unlink(missing_ok=True)
        shutil.move(reqs_file, reqs_file_backup)
        try:
            self.build_dependencies()
            original_requirements = "\n".join(
                sorted(reqs_file_backup.read_text().splitlines())
            )
            new_requirements = "\n".join(sorted(reqs_file.read_text().splitlines()))
            if original_requirements != new_requirements:
                raise RuntimeError(
                    f"""Requirement files do not match. Original requirements:
    {original_requirements}
    
    Newly built requirements:
    {new_requirements}
                """
                )
        finally:
            reqs_file.unlink(missing_ok=True)
            shutil.move(reqs_file_backup, reqs_file)

    def generate_requirements(self, extras):
        pdm_file = {
            "project": {
                "name": f"{self.config.section}-repository",
                "version": "1.0.0",
                "description": "",
                "packages": [],
                "authors": [
                    {
                        "name": self.config["author_name"],
                        "email": self.config["author_email"],
                    },
                ],
                "dependencies": [OAREPO_VERSION, *extras],
                "requires-python": PYTHON_VERSION,
            }
        }
        with open(self.site_dir / "pyproject.toml", "wb") as f:
            tomli_w.dump(pdm_file, f)

        self.call_pdm(
            "lock",
        )
        return self.call_pdm(
            "export", "-f", "requirements", "--without-hashes", grab_stdout=True
        )

    def init_files(self):
        host, port, access_key, secret_key = self.get_invenio_configuration(
            "INVENIO_S3_HOST",
            "INVENIO_S3_PORT",
            "INVENIO_S3_ACCESS_KEY",
            "INVENIO_S3_SECRET_KEY",
        )

        client = Minio(
            f"{host}:{port}",
            access_key=access_key,
            secret_key=secret_key,
            secure=False,
        )

        bucket_name = self.config["site_package"].replace(
            "_", ""
        )  # bucket names with underscores are not allowed
        if not client.bucket_exists(bucket_name):
            client.make_bucket(bucket_name)
        self.call_invenio(
            "files", "location", "default", f"s3://{bucket_name}", "--default"
        )
        self.check_file_location_initialized(raise_error=True)

    def check_file_location_initialized(self, raise_error=False):
        try:
            output = self.call_invenio(
                "files",
                "location",
                "list",
                grab_stdout=True,
                raise_exception=True,
            )
            print(f"initialization check:\n{output}\n")
        except subprocess.CalledProcessError:
            raise Exception("Checking if file location exists failed.")
        if output:
            return True
        else:
            if raise_error:
                raise Exception(
                    "No file location exists. This probably means that the wizard was unable to create one."
                )
            return False

    def get_invenio_configuration(self, *keys):
        values = dotenv_values(self.site_dir / "variables")
        values.update({k: v for k, v in os.environ.items() if k.startswith("INVENIO_")})

        def convert(x):
            try:
                if x == "False":
                    return False
                if x == "True":
                    return True
                return json.loads(x)
            except:
                return x

        try:
            return [convert(values[x]) for x in keys]
        except KeyError as e:
            raise KeyError(
                f"Configuration key not found in defaults: {values.keys()}: {e}"
            )



from oarepo_cli.wizard import Wizard
from oarepo_cli.wizard.docker import DockerRunner

from ..site.add.steps.link_env import LinkEnvStep
from ..site.add.steps.start_containers import StartContainersStep
from .steps.check_db import CheckDBStep
from .steps.check_dependencies import CheckDependenciesStep
from .steps.check_s3_location import CheckS3LocationStep
from .steps.check_search import CheckSearchStep
from .steps.check_site import CheckSiteStep
from .steps.check_ui import CheckUIStep
from .steps.check_venv import CheckVirtualenvStep
from .steps.develop_step import DevelopStep
from .steps.editor_support import EditorSupportStep


class DevelopWizard(Wizard):
    def __init__(self, runner: DockerRunner, *, site_support):
        self.site_support = site_support
        super().__init__(
            LinkEnvStep(),
            StartContainersStep(),
            *runner.wrap_docker_steps(
                CheckVirtualenvStep(),
                CheckDependenciesStep(),
                CheckSiteStep(),
                CheckUIStep(),
                CheckDBStep(),
                CheckSearchStep(),
                CheckS3LocationStep(),
                EditorSupportStep(),
            ),
            DevelopStep(),
        )



from __future__ import annotations

import os

from oarepo_cli.site.mixins import SiteWizardStepMixin
from oarepo_cli.utils import ProjectWizardMixin
from oarepo_cli.wizard import WizardStep


class LinkEnvStep(SiteWizardStepMixin, ProjectWizardMixin, WizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            heading="""
I link the "variables" file in the site directory to the .env file. 
If you'd like to make local changes to the variables, remove the link,
"cp variables .env" and edit the file.""",
            **kwargs,
        )

    def after_run(self):
        try:
            # try to unlink stale symlink - it also does not "exist" from pathlib point of view
            os.unlink(str(self.site_dir / ".env"))
        except:
            pass
        os.symlink("variables", str(self.site_dir / ".env"))

    def should_run(self):
        return not (self.site_dir / ".env").exists()



import venv
from pathlib import Path

import yaml

from oarepo_cli.model.utils import ModelWizardStep
from oarepo_cli.package_versions import OAREPO_MODEL_BUILDER_VERSION
from oarepo_cli.utils import pip_install, run_cmdline
from oarepo_cli.wizard import WizardStep


class CompileWizardStep(ModelWizardStep, WizardStep):
    def after_run(self):
        if self.data.running_in_docker:
            venv_dir = Path("/tmp/oarepo-model-builder")
        else:
            venv_dir = (
                self.data.project_dir
                / ".venv"
                / f"oarepo-model-builder-{self.model_name}"
            )
        print("model builder venv dir", venv_dir)
        venv_dir = venv_dir.absolute()
        if not venv_dir.exists():
            venv_args = [str(venv_dir)]
            if self.data.running_in_docker:
                venv_args.append("--system-site-packages")
            venv.main(venv_args)

        run_cmdline(
            venv_dir / "bin" / "pip",
            "list",
            cwd=self.model_dir,
        )

        pip_install(
            venv_dir / "bin" / "pip",
            "OAREPO_MODEL_BUILDER_VERSION",
            f"oarepo-model-builder{OAREPO_MODEL_BUILDER_VERSION}",
            "https://github.com/oarepo/oarepo-model-builder",
        )

        with open(self.model_dir / "model.yaml") as f:
            model_data = yaml.safe_load(f)

        # install plugins from model.yaml
        self._install_plugins_from_model(model_data, venv_dir)

        # install plugins from included files
        uses = model_data.get("use") or []
        if not isinstance(uses, list):
            uses = [uses]

        for use in uses:
            if not use.startswith("."):
                # can not currently find plugins in uses
                # that are registered as entrypoints
                continue
            with open(self.model_dir.joinpath(use)) as f:
                used_data = yaml.safe_load(f)
                self._install_plugins_from_model(used_data, venv_dir)

        run_cmdline(
            venv_dir / "bin" / "oarepo-compile-model",
            "-vvv",
            "model.yaml",
            cwd=self.model_dir,
        )

    def _install_plugins_from_model(self, model_data, venv_dir):
        plugins = model_data.get("plugins", {}).get("packages", [])
        for package in plugins:
            run_cmdline(
                venv_dir / "bin" / "pip",
                "install",
                package,
                cwd=self.model_dir,
            )

    def should_run(self):
        # always run as there is an optional step for merge/overwrite changes
        return True



"""JS/CSS Webpack bundles for {{cookiecutter.project_name}}."""

from invenio_assets.webpack import WebpackThemeBundle

theme = WebpackThemeBundle(
    __name__,
    "assets",
    default="semantic-ui",
    themes={
        "semantic-ui": dict(
            entry={
                # Add your webpack entrypoints
            },
            devDependencies={
                "eslint": "^8.0.0",
                "eslint-config-react-app": "^7.0.0",
                "prettier": "^2.8.0",
                "eslint-config-prettier": "^8.8.0",
                "@typescript-eslint/eslint-plugin": "^5.0.0",
                "@typescript-eslint/parser": "^5.0.0",
                "typescript": "^5.0.0",
                "@babel/plugin-proposal-private-property-in-object": "^7.0.0",
            },
            aliases={},
        ),
    },
)



import os

from oarepo_cli.model.model_support import ModelSupport
from oarepo_cli.site.site_support import SiteSupport
from oarepo_cli.utils import ProjectWizardMixin, SiteMixin
from oarepo_cli.wizard import WizardStep


class ModelWizardStep(SiteMixin, ProjectWizardMixin, WizardStep):
    @property
    def model_support(self) -> ModelSupport:
        return self.root.model_support

    @property
    def site_support(self) -> SiteSupport:
        return self.root.site_support

    @property
    def model_name(self):
        return self.model_support.model_name

    @property
    def model_dir(self):
        return self.model_support.model_dir

    @property
    def model_package_dir(self):
        return self.model_dir / os.sep.join(self.data["model_package"].split("."))



from oarepo_cli.model.compile.steps.compile import CompileWizardStep
from oarepo_cli.model.compile.steps.remove_previous import RemovePreviousModelStep
from oarepo_cli.wizard import Wizard
from oarepo_cli.wizard.docker import DockerRunner


class CompileModelWizard(Wizard):
    def __init__(self, runner: DockerRunner):
        super().__init__(
            RemovePreviousModelStep(),
            *runner.wrap_docker_steps(CompileWizardStep(), in_compose=False)
        )



from collections import defaultdict

from .base import BaseFormatStep


class FormatPythonStep(BaseFormatStep):
    def format_paths(self, model_paths, ui_paths, local_paths, site_paths):
        python_source_paths = [*model_paths, *ui_paths, *local_paths, *site_paths]
        self.run_autoflake(python_source_paths, exclude=[])
        self.run_isort(python_source_paths)
        self.run_black(python_source_paths)

    def run_autoflake(self, dirs, exclude):
        from autoflake import find_files, fix_file

        files = list(find_files([*dirs], True, exclude))

        for file_name in files:
            try:
                fix_file(
                    file_name,
                    args=defaultdict(
                        lambda: None,
                        in_place=True,
                        remove_all_unused_imports=True,
                        write_to_stdout=False,
                        verbose=True,
                    ),
                )
            except Exception as e:
                print(f"Error in autoflaking {file_name}: {e}")

    def run_isort(self, dirs):
        from isort.main import main

        main(["--profile", "black", *[str(x) for x in dirs]])

    def run_black(self, dirs):
        import black

        black.main([*[str(x) for x in dirs]], standalone_mode=False)



{{cookiecutter.app_package|upper}}_RESOURCE = "{{cookiecutter.app_package}}.resources.resource:{{cookiecutter.resource}}"
{{cookiecutter.app_package|upper}}_RESOURCE_CONFIG = "{{cookiecutter.app_package}}.resources.config:{{cookiecutter.resource_config}}"



import json


class UIWizardMixin:
    @property
    def ui_name(self):
        return self.data.section

    @property
    def ui_dir(self):
        return self.data.project_dir / "ui" / self.ui_name


class AssociatedModelMixin:
    def get_model_definition(self):
        model_config = self.data.whole_data["models"][self.data["model_name"]]
        model_package = model_config["model_package"]

        model_path = self.data.project_dir / model_config["model_dir"]
        model_file = (
            (model_path / model_package / "models" / "records.json")
            .absolute()
            .resolve(strict=False)
        )
        with open(model_file) as f:
            model_description = json.load(f)
        return model_description, model_path, model_package, model_config



from oarepo_cli.invenio.steps.run_invenio import RunInvenioStep
from oarepo_cli.wizard import Wizard
from oarepo_cli.wizard.docker import DockerRunner


class InvenioWizard(Wizard):
    def __init__(self, runner: DockerRunner, *, site_support, invenio_args):
        self.site_support = site_support

        super().__init__(
            *runner.wrap_docker_steps(
                RunInvenioStep(invenio_args),
            ),
        )



from oarepo_cli.model.gen.base import GeneratedFile
from oarepo_cli.model.utils import ModelWizardStep
from oarepo_cli.utils import unique_merger
from oarepo_cli.wizard import RadioStep


class AddMetadataWizardStep(ModelWizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            RadioStep(
                "use_metadata",
                default="yes",
                heading="Do you want to define your own metadata in a separated 'metadata' element and file ?",
                options={
                    "yes": "Yes, I want metadata separated in a new file",
                    "no": "No, I'll define metadata directly in the model file",
                },
            ),
            **kwargs
        )

    def after_run(self):
        if self.data["use_metadata"] != "yes":
            return

        yaml_file: GeneratedFile = self.root.files.get("model.yaml")
        yaml = yaml_file.yaml
        unique_merger.merge(
            yaml,
            {
                "record": {"properties": {"use": ["./metadata.yaml"]}},
            },
        )
        yaml_file.save()

        yaml_file: GeneratedFile = self.root.files.get("metadata.yaml")
        yaml = yaml_file.yaml
        unique_merger.merge(yaml, {"metadata": {"properties": {}}})
        yaml_file.save()

    def should_run(self):
        return True



from oarepo_cli.watch.steps.watcher_step import WatcherStep
from oarepo_cli.wizard import Wizard
from oarepo_cli.wizard.docker import DockerRunner


class WatcherWizard(Wizard):
    def __init__(self, runner: DockerRunner, site_support, run_ui):
        self.site_support = site_support
        super().__init__(*runner.wrap_docker_steps(WatcherStep(run_ui)))



import json
import os
import re
import shutil
from pathlib import Path

from oarepo_cli.site.site_support import SiteSupport
from oarepo_cli.utils import copy_tree
from oarepo_cli.wizard import WizardStep


class EditorSupportStep(WizardStep):
    def should_run(self):
        return True

    def after_run(self):
        # create package.json
        site_support: SiteSupport = self.root.site_support

        (self.data.project_dir / "package.json").write_text(
            json.dumps(
                {
                    "name": "for-editor-autocompletion",
                    "description": re.sub(
                        r"\n\s*",
                        " ",
                        """
                    This is an autogenerated package to enable editor autocompletion. 
                    It is not supposed to be committed into git and might be removed at any time.
              """,
                    ).strip(),
                    "version": "1.0.0",
                },
                indent=4,
            )
        )

        # create tsconfig.json

        site_support.call_invenio(
            "oarepo",
            "assets",
            "collect",
            f"{site_support.invenio_instance_path}/watch.list.json",
            "--repository-dir",
            str(self.data.project_dir),
            "--assets-dir",
            "./.assets",
        )

        watch_list = json.loads(
            (site_support.invenio_instance_path / "watch.list.json").read_text()
        )

        aliases = {}
        for alias, root_path in watch_list.get("@root_aliases", {}).items():
            if not root_path.startswith("./"):
                root_path = "./" + root_path
            aliases.setdefault(alias, []).append(root_path)

        Path(self.data.project_dir / "jsconfig.json").write_text(
            json.dumps(
                {"baseUrl": ".", "compilerOptions": {"paths": aliases}},
                indent=4,
                ensure_ascii=False,
            )
        )

        # copy or link node_modules
        assets_path: Path = self.data.project_dir / ".assets"
        modules_path: Path = self.data.project_dir / "node_modules"

        try:
            # if is link, unlink it
            if assets_path.is_dir() and not assets_path.is_symlink():
                shutil.rmtree(assets_path)
        except Exception as e:
            print(
                f"Could not remove {assets_path} as a directory, will try to remove it as a link: {e}"
            )

        try:
            assets_path.unlink()
        except Exception as e:
            print(
                f"Could not remove {assets_path} as a file/link, might have been already removed: {e}"
            )

        try:
            modules_path.unlink(missing_ok=True)
        except Exception as e:
            print(
                f"Could not remove {modules_path} as a file/link, might have been already removed: {e}"
            )

        if self.data.running_in_docker:
            copy_tree(site_support.invenio_instance_path / "assets", assets_path)
        else:
            assets_path.symlink_to(site_support.invenio_instance_path / "assets")

        # can not use Pathlib.symlink_to as it does not create a relative symlink
        os.symlink(".assets/node_modules", modules_path)



import functools
import os
import re
import select
import shutil
import subprocess
import sys
from pathlib import Path

import click
import deepmerge
import git
import pydriller
import yaml
from colorama import Fore, Style

from oarepo_cli.config import MonorepoConfig


def run_cmdline(
    *cmdline,
    cwd=".",
    environ=None,
    check_only=False,
    grab_stdout=False,
    grab_stderr=False,
    discard_output=False,
    raise_exception=False,
    with_tty=False,
    no_input=False,
    no_environment=False,
):
    if no_environment:
        env = {}
    else:
        env = os.environ.copy()
        env.update(environ or {})
    cwd = Path(cwd).absolute()
    cmdline = [str(x) for x in cmdline]
    print(
        f"{Fore.BLUE}Running {Style.RESET_ALL} {' '.join(cmdline)}", file=sys.__stderr__
    )
    if "DOCKER_AROUND" in os.environ:
        inside_docker = "docker at path "
    else:
        inside_docker = ""
    print(
        f"{Fore.BLUE}    inside {Style.RESET_ALL} {inside_docker}{cwd}",
        file=sys.__stderr__,
    )
    try:
        kwargs = {}
        if no_input:
            kwargs["stdin"] = subprocess.DEVNULL
        if grab_stdout or grab_stderr or discard_output:
            if grab_stdout or discard_output:
                kwargs["stdout"] = subprocess.PIPE
            if grab_stderr or discard_output:
                kwargs["stderr"] = subprocess.PIPE

            ret = subprocess.run(
                cmdline,
                check=True,
                cwd=cwd,
                env=env,
                **kwargs,
            )
            ret = (ret.stdout or b"") + b"\n" + (ret.stderr or b"")
        else:
            if with_tty:
                ret = run_with_tty(cmdline, cwd=cwd, env=env)
            else:
                ret = subprocess.call(cmdline, cwd=cwd, env=env, **kwargs)
            if ret:
                raise subprocess.CalledProcessError(ret, cmdline)
    except subprocess.CalledProcessError as e:
        if check_only:
            return False
        print(f"Error running {' '.join(cmdline)}", file=sys.__stderr__)
        if e.stdout:
            print(e.stdout)
        if e.stderr:
            print(e.stderr)
        if raise_exception:
            raise
        sys.exit(e.returncode)
    print(
        f"{Fore.GREEN}Finished running {Style.RESET_ALL} {' '.join(cmdline)}",
        file=sys.__stderr__,
    )
    print(f"{Fore.GREEN}    inside {Style.RESET_ALL} {cwd}", file=sys.__stderr__)
    if grab_stdout:
        return ret.decode("utf-8").strip()
    return True


def run_with_tty(cmd, cwd=None, env=None):
    # https://stackoverflow.com/questions/41542960/run-interactive-bash-with-popen-and-a-dedicated-tty-python
    import pty
    import termios
    import tty

    # save original tty setting then set it to raw mode
    old_tty = termios.tcgetattr(sys.stdin)
    tty.setraw(sys.stdin.fileno())

    # open pseudo-terminal to interact with subprocess
    master_fd, slave_fd = pty.openpty()
    try:
        # use os.setsid() make it run in a new process group, or bash job control will not be enabled
        p = subprocess.Popen(
            cmd,
            preexec_fn=os.setsid,
            stdin=slave_fd,
            stdout=slave_fd,
            stderr=slave_fd,
            universal_newlines=True,
        )

        while p.poll() is None:
            r, w, e = select.select([sys.stdin, master_fd], [], [], 1)
            if sys.stdin in r:
                d = os.read(sys.stdin.fileno(), 10240)
                os.write(master_fd, d)
            elif master_fd in r:
                o = os.read(master_fd, 10240)
                if o:
                    os.write(sys.stdout.fileno(), o)
        return p.returncode
    finally:
        # restore tty settings back
        termios.tcsetattr(sys.stdin, termios.TCSADRAIN, old_tty)


def find_oarepo_project(dirname, raises=False):
    dirname = Path(dirname).absolute()
    orig_dirname = dirname
    for _ in range(4):
        if (dirname / "oarepo.yaml").exists():
            return dirname
        dirname = dirname.parent
    if raises:
        raise Exception(
            f"Not part of OARepo project: directory {orig_dirname} "
            f"or its 4 ancestors do not contain oarepo.yaml file"
        )
    return


def to_python_name(x):
    x = re.sub(r"(?<!^)(?=[A-Z])", "_", x).lower()
    x = x.replace("-", "_")
    return re.sub("[^a-z_]", "", x)


def pip_install(pip_binary, env_name, lib_name_and_version, lib_github):
    # run pip installation, taking env versions into account
    run_cmdline(pip_binary, "install", "-U", "--no-input", "setuptools", "pip", "wheel")
    installation_option = os.environ.get(env_name, "release")
    if installation_option == "release":
        # release
        run_cmdline(pip_binary, "install", "--no-input", lib_name_and_version)
    elif installation_option == "maintrunk":
        run_cmdline(
            pip_binary,
            "install",
            "--no-input",
            f"git+{lib_github}",
        )
    elif installation_option.startswith("https://"):
        run_cmdline(
            pip_binary,
            "install",
            "--no-input",
            f"git+{installation_option}",
        )
    else:
        run_cmdline(
            pip_binary,
            "install",
            "--no-input",
            "-e",
            Path(installation_option),
        )


def get_cookiecutter_source(env_name, lib_github, lib_version, master_version="master"):
    installation_option = os.environ.get(env_name, "release")
    if installation_option == "release":
        cookiecutter_path = lib_github
        cookiecutter_branch = lib_version
    elif installation_option == "maintrunk":
        cookiecutter_path = lib_github
        cookiecutter_branch = master_version
    elif installation_option.startswith("https://"):
        # something like https://github.com/oarepo/oarepo-model-builder/tree/datatypes
        cookiecutter_path, cookiecutter_branch = installation_option.rsplit(
            "/tree/", maxsplit=1
        )
    else:
        cookiecutter_path = installation_option
        cookiecutter_branch = None
    return cookiecutter_path, cookiecutter_branch


def commit_git(repo_dir, tag_name, message):
    if "DOCKER_AROUND" in os.environ:
        return
    tag_index = 1

    if not (repo_dir / ".git").exists():
        git.Repo.init(repo_dir)

    try:
        for commit in pydriller.Repository(str(repo_dir)).traverse_commits():
            for m in commit.msg.split("\n"):
                if m.startswith("omb-"):
                    tag_index += 1
    except git.exc.GitCommandError:
        pass
    except git.exc.InvalidGitRepositoryError:
        pass

    repo = git.Repo(repo_dir)
    tag_name = f"omb-{tag_index:05d}-{tag_name}"
    repo.git.add(repo_dir)
    index = repo.index
    if index.entries:
        index.commit(message + "\n\n" + tag_name)


def must_be_committed(repo_dir):
    if "DOCKER_AROUND" in os.environ:
        return
    repo = git.Repo(repo_dir)
    if repo.is_dirty() or repo.untracked_files:
        for f in repo.untracked_files:
            print("    ", f)

        print(
            "The repository contains untracked or dirty files. Please commit/ignore them before continuing."
        )
        sys.exit(1)


def copy_tree(src, dest):
    to_copy = [(src, dest)]
    copied_files = []
    while to_copy:
        source, destination = to_copy.pop()
        if os.path.isdir(source):
            print(f"Copying directory {source} -> {destination}")
            if os.path.exists(destination):
                print("    ... already exists")
                if not os.path.isdir(destination):
                    raise AttributeError(
                        f"Destination {destination} should be a directory but is {path_type(destination)}"
                    )
            else:
                print("    ... creating and testing directory")
                os.makedirs(destination)
                if not os.path.isdir(destination):
                    raise AttributeError(
                        f"I've just created a {destination} directory but it failed and I've got {path_type(destination)}"
                    )
            for fn in reversed(os.listdir(source)):
                to_copy.append(
                    (os.path.join(source, fn), os.path.join(destination, fn))
                )
        else:
            print(f"Copying file {source} -> {destination}")
            if os.path.exists(destination):
                os.unlink(destination)
            if os.path.exists(destination):
                raise AttributeError(
                    f"I've just deleted {destination}, but it still exists and is {path_type(destination)}"
                )

            shutil.copy(source, destination, follow_symlinks=True)
            if not os.path.isfile(destination):
                raise AttributeError(
                    f"I've just copied file {source} into {destination}, but the destination is not a file, it is {path_type(destination)}"
                )
            if (
                os.stat(source, follow_symlinks=True).st_size
                != os.stat(destination).st_size
            ):
                raise AttributeError(
                    f"I've just copied file {source} into {destination}, but the sizes do not match. "
                    f"Source size {os.stat(source).st_size}, destination size {os.stat(destination).st_size}"
                )
            copied_files.append(destination)
    return copied_files


def path_type(path):
    if os.path.isdir(path):
        return "dir"
    elif os.path.isfile(path):
        return "file"
    elif os.path.islink(path):
        return "link"
    else:
        return "unknown"


unique_merger = deepmerge.Merger(
    [(list, ["append_unique"]), (dict, ["merge"]), (set, ["union"])],
    # next, choose the fallback strategies,
    # applied to all other types:
    ["override"],
    # finally, choose the strategies in
    # the case where the types conflict:
    ["override"],
)


def snail_to_title(v):
    return "".join(ele.title() for ele in v.split("_"))


def with_config(
    config_section=None, project_dir_as_argument=False, config_as_argument=False
):
    def wrapper(f):
        @(
            click.argument(
                "project_dir",
                type=click.Path(exists=False, file_okay=False),
                required=True,
            )
            if project_dir_as_argument
            else click.option(
                "--project-dir",
                type=click.Path(exists=True, file_okay=False),
                required=False,
                help="Directory containing an already initialized project. "
                "If not set, current directory is used",
            )
        )
        @click.option(
            "--no-input",
            is_flag=True,
            type=bool,
            required=False,
            help="Take options from the config file, skip user input",
        )
        @click.option(
            "--silent",
            is_flag=True,
            type=bool,
            required=False,
            help="Do not output program's messages. "
            "External program messages will still be displayed",
        )
        @click.option(
            "--verbose",
            is_flag=True,
            type=bool,
            required=False,
            help="Verbose output",
        )
        @click.option(
            "--step",
            required=False,
            multiple=True,
            help="Run only this step",
        )
        @click.option(
            "--steps",
            is_flag=True,
            type=bool,
            required=False,
            help="List all steps and exit",
        )
        @(
            click.argument(
                "config",
                type=click.Path(exists=True, file_okay=True, dir_okay=False),
                required=True,
            )
            if config_as_argument
            else click.option(
                "--config",
                type=click.Path(exists=True, file_okay=True, dir_okay=False),
                required=False,
                help="Merge this config to the main config in target directory and proceed",
            )
        )
        @click.option(
            "--use-docker",
            "use_docker",
            flag_value="docker",
            default=None,
            help="Run the command inside docker",
        )
        @click.option(
            "--outside-docker",
            "use_docker",
            flag_value="no docker",
            help="Run the command outside of docker container",
        )
        @click.pass_context
        @functools.wraps(f)
        def wrapped(
            context,
            project_dir=None,
            config=None,
            use_docker=None,
            **kwargs,
        ):
            if not project_dir:
                project_dir = Path.cwd()
            project_dir = Path(project_dir).absolute()
            oarepo_yaml_file = project_dir / "oarepo.yaml"

            if callable(config_section):
                section = config_section(**kwargs)
            else:
                section = config_section or "config"

            cfg = MonorepoConfig(oarepo_yaml_file, section=section)

            if oarepo_yaml_file.exists():
                cfg.load()

            if config:
                config_data = yaml.safe_load(Path(config).read_text())
                cfg.merge_config(config_data, top=not config_section)

            kwargs.pop("cfg", None)
            kwargs.pop("project_dir", None)

            cfg.running_in_docker = "DOCKER_AROUND" in os.environ
            if not use_docker:
                cfg.use_docker = None
            else:
                cfg.use_docker = use_docker == "docker"

            cfg.no_input = kwargs.get("no_input", False)

            try:
                return f(context=context, project_dir=project_dir, cfg=cfg, **kwargs)
            except Exception as e:
                if kwargs.get("verbose"):
                    import traceback

                    traceback.print_exc()
                else:
                    print(str(e))
                raise

        return wrapped

    return wrapper


def load_user_config(project_dir):
    if not project_dir.exists():
        return {}

    local_user_config = {}
    local_user_config_file = project_dir / ".oarepo-local.yaml"
    if local_user_config_file.exists():
        with open(local_user_config_file) as f:
            local_user_config = yaml.safe_load(f)

    if "use_docker" not in local_user_config:
        click.secho(
            """I can isolate all the commands into a docker container.
    The advantage is that only python and docker are needed, disadvantage is a slight performance
    penalty.
    
    Should I do so? (y/n)
            """
        )
        yn = input().strip().lower()
        if yn == "y":
            local_user_config["use_docker"] = True
            with open(local_user_config_file, "w") as f:
                yaml.safe_dump(local_user_config, f)
    return local_user_config


class ProjectWizardMixin:
    @property
    def site_dir(self):
        if not hasattr(self, "site"):
            raise Exception("Current site not set")
        return self.data.project_dir / self.site["site_dir"]

    @property
    def oarepo_cli(self):
        return self.data.project_dir / self.data.get("config.oarepo_cli")

    def run_cookiecutter(
        self,
        template,
        config_file,
        checkout=None,
        output_dir=None,
        extra_context=None,
        environ=None,
    ):
        config_dir: Path = self.data.project_dir / ".cookiecutters"
        config_dir.mkdir(parents=True, exist_ok=True)
        config_file = config_dir / f"{config_file}.yaml"
        config_file.write_text(yaml.safe_dump({"default_context": extra_context}))
        cookiecutter_command = (
            Path(sys.executable or sys.argv[0]).absolute().parent / "cookiecutter"
        )
        output_dir_temp = f"{output_dir}-tmp"
        output_dir = Path(output_dir)
        output_dir_temp = Path(output_dir_temp)
        args = [
            template,
            "--no-input",
            "-o",
            output_dir_temp,
            "--config-file",
            config_file,
        ]
        if checkout:
            args.append("-c")
            args.append(checkout)

        run_cmdline(
            cookiecutter_command,
            *args,
            cwd=self.data.project_dir,
            environ={**(environ or {})},
        )
        merge_from_temp_to_target(output_dir_temp, output_dir)


class SiteMixin(ProjectWizardMixin):
    @property
    def site_dir(self):
        site_name = self.data.get("sites", [])
        if not site_name:
            raise Exception("Unexpected error: No installation site specified")
        site = self.data.get(f"sites.{site_name[0]}")
        if not site:
            raise Exception(
                f"Unexpected error: Site with name {site_name[0]} does not exist"
            )
        return self.data.project_dir / site["site_dir"]


def merge_from_temp_to_target(output_dir_temp, output_dir):
    source: Path
    for source in output_dir_temp.rglob("*"):
        rel = source.relative_to(output_dir_temp)
        dest: Path = output_dir / rel
        if source.is_dir():
            if not dest.exists():
                dest.mkdir(parents=True)
        elif source.is_file():
            if not dest.exists():
                if not dest.parent.exists():
                    dest.parent.mkdir(parents=True)
                shutil.copy(source, dest)
    shutil.rmtree(output_dir_temp)


def check_call(*args, **kwargs):
    cmdline = " ".join(str(x) for x in args[0])
    print(f"Calling command {cmdline} with kwargs {kwargs}")
    return subprocess.check_call(*args, **kwargs)


def run_nrp_in_docker_compose(
    site_dir, *arguments, interactive=True, no_input=False, networking=True, name=None
):
    run_cmdline(
        "docker",
        "compose",
        "run",
        *(["--service-ports"] if networking else []),
        "--rm",
        *(["-i"] if interactive else []),
        *(["--no-TTY"] if no_input else []),
        *(["--name", name] if name else []),
        "repo",
        *arguments,
        cwd=site_dir,
        environ={**os.environ, "INVENIO_DOCKER_USER_ID": str(os.getuid())},
        with_tty=False,
        no_input=no_input,
    )


def exec_nrp_in_docker(site_dir, container_name, *arguments, interactive=True):
    run_cmdline(
        "docker",
        "exec",
        *(["-it"] if interactive else []),
        container_name,
        "/nrp/bin/nrp",
        *arguments,
        cwd=site_dir,
        environ={**os.environ, "INVENIO_DOCKER_USER_ID": str(os.getuid())},
        with_tty=False,
        no_input=not interactive,
    )


def run_nrp_in_docker(repo_dir: Path, *arguments, interactive=True):
    run_cmdline(
        "docker",
        "run",
        *(["-it"] if interactive else []),
        "-v",
        f"{str(repo_dir)}:/repository",
        "--user",
        f"{os.getuid()}:{os.getgid()}",
        "-e",
        f"REPOSITORY_DIR={repo_dir.name}",
        "--rm",
        "oarepo/oarepo-base-development:11",
        *arguments,
        cwd=repo_dir,
        with_tty=True,
    )


def batched(lst, n):
    if n < 1:
        raise ValueError("n must be at least one")

    while lst:
        data = lst[:n]
        lst = lst[n:]
        yield data



from oarepo_cli.wizard import StaticStep, Wizard

from ...wizard.docker import DockerRunner
from .steps.add_ui import AddUIWizardStep
from .steps.jinja import CreateJinjaStep


class AddUIWizard(Wizard):
    def __init__(self, runner: DockerRunner, *, site_support):
        self.site_support = site_support
        super().__init__(
            StaticStep(
                heading="""
        A UI is a python package that displays the search, detail, edit, ... pages for a single
        metadata model. At first you'll have to select the model for which the UI will be created
        and then I'll ask you a couple of additional questions.
        """,
            ),
            AddUIWizardStep(),
            *runner.wrap_docker_steps(
                CreateJinjaStep(),
            )
        )



from oarepo_cli.upgrade.steps.upgrade_docker_nrp import UpgradeDockerNRPStep
from oarepo_cli.upgrade.steps.upgrade_nrp import UpgradeNRPStep
from oarepo_cli.upgrade.steps.upgrade_site import UpgradeSiteStep
from oarepo_cli.wizard import Wizard
from oarepo_cli.wizard.docker import DockerRunner


class UpgradeWizard(Wizard):
    def __init__(self, runner: DockerRunner, sites=None, site=None):
        upgrade_sites_steps = []
        for s in sites:
            if site and s != site:
                continue
            upgrade_sites_steps.extend(
                runner.wrap_docker_steps(
                    UpgradeDockerNRPStep(), UpgradeSiteStep(s), site=s
                )
            )

        super().__init__(
            # upgrade nrp tooling outside of docker
            UpgradeNRPStep(),
            *upgrade_sites_steps
        )



from __future__ import annotations

import shutil
import venv

from oarepo_cli.package_versions import NRP_CLI_VERSION
from oarepo_cli.utils import commit_git, pip_install
from oarepo_cli.wizard import WizardStep


class InstallINRPCliStep(WizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            heading="""
I will install nrp command-line tools that make using the invenio easier.
To run them, invoke the "nrp" script from within the project directory.            
            """,
            **kwargs,
        )

    def after_run(self):
        print("Creating nrp virtualenv")
        self.data["oarepo_cli"] = str(
            (self.nrp_cli_dir / "bin" / "nrp").relative_to(self.data.project_dir)
        )
        if self.nrp_cli_dir.exists():
            shutil.rmtree(self.nrp_cli_dir)
        venv.main([str(self.nrp_cli_dir)])

        pip_install(
            self.nrp_cli_dir / "bin" / "pip",
            "NRP_CLI_VERSION",
            NRP_CLI_VERSION,
            "https://github.com/oarepo/oarepo-cli",
        )

        with open(self.nrp_cli_dir / ".check.ok", "w") as f:
            f.write("oarepo check ok")
        commit_git(
            self.data.project_dir,
            "after-install-oarepo-cli",
            "Committed automatically after oarepo-cli has been installed",
        )

    @property
    def nrp_cli_dir(self):
        return self.data.project_dir / ".venv" / "nrp"

    def should_run(self):
        return not (self.nrp_cli_dir / ".check.ok").exists()



from oarepo_cli.wizard import Wizard

from ...wizard.docker import DockerRunner
from .steps.uninstall import UnInstallLocalStep


class RemoveLocalWizard(Wizard):
    def __init__(self, runner: DockerRunner, *, site_support):
        self.site_support = site_support
        super().__init__(*runner.wrap_docker_steps(UnInstallLocalStep()))



from oarepo_cli.model.gen.base import GeneratedFile
from oarepo_cli.model.utils import ModelWizardStep
from oarepo_cli.package_versions import OAREPO_MODEL_BUILDER_FILES_VERSION
from oarepo_cli.utils import unique_merger
from oarepo_cli.wizard import RadioStep


class AddFilesWizardStep(ModelWizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            RadioStep(
                "use_files",
                heading="Do you want to store digital objects along with metadata?",
                options={
                    "yes": "Yes, I need files",
                    "no": "No, this is metadata-only repository",
                },
            ),
            **kwargs,
        )

    def after_run(self):
        if self.data["use_files"] != "yes":
            return

        yaml_file: GeneratedFile = self.root.files.get("model.yaml")
        yaml = yaml_file.yaml
        unique_merger.merge(
            yaml,
            {
                "record": {"use": ["./files.yaml"]},
                "plugins": {
                    "packages": [
                        f"oarepo-model-builder-files{OAREPO_MODEL_BUILDER_FILES_VERSION}"
                    ]
                },
                "profiles": ["files"],
            },
        )
        yaml_file.save()

        yaml_file: GeneratedFile = self.root.files.get("files.yaml")
        yaml = yaml_file.yaml
        unique_merger.merge(
            yaml,
            {
                "files": {
                    "use": ["invenio_files"],
                    "module": {
                        "qualified": self.data["model_package"],
                    },
                    "properties": {},
                }
            },
        )
        yaml_file.save()

    def should_run(self):
        return True



import click as click

from oarepo_cli.utils import with_config


@click.command(name="list", help="List installed models")
@with_config()
def list_models(cfg=None, **kwargs):
    for model in cfg.whole_data.get("models", {}):
        print(model)



from oarepo_cli.site.mixins import SiteWizardStepMixin
from oarepo_cli.wizard import WizardStep


class InitFilesStep(SiteWizardStepMixin, WizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            heading="""
        Now I will configure the default location for files storage in the minio s3 framework.
            """,
            **kwargs,
        )

    def after_run(self):
        self.site_support.init_files()

    def should_run(self):
        return not self.site_support.check_file_location_initialized(raise_error=False)



import typing
from configparser import ConfigParser
from io import StringIO

from ruamel.yaml import YAML, CommentedMap

from oarepo_cli.model.gen.base import FileFormat


class TextFormat(FileFormat[str]):
    def load(self, text: str) -> str:
        return text or ""

    def dump(self, data: str) -> str:
        return data


class YAMLFormat(FileFormat[typing.Any]):
    def __init__(self):
        self._yaml = YAML(typ="rt")

    def load(self, text: str) -> typing.Any:
        return self._yaml.load(StringIO(text)) or CommentedMap()

    def dump(self, data: typing.Any) -> str:
        io = StringIO()
        self._yaml.dump(data, io)
        return io.getvalue()


class PermissiveConfigParser(ConfigParser):
    def get_section(self, section_name):
        if not self.has_section(section_name):
            self.add_section(section_name)
        return self[section_name]


class CFGFormat(FileFormat[ConfigParser]):
    def load(self, text: str) -> ConfigParser:
        parser = ConfigParser()
        parser.read_string(text)
        return parser

    def dump(self, data: ConfigParser) -> str:
        io = StringIO()
        data.write(io)
        return io.getvalue()



import json
import os
import re

from colorama import Fore, Style

from oarepo_cli.model.utils import ModelWizardStep


class CreateAlembicModelStep(ModelWizardStep):
    heading = f"""
    I will create/update the alembic migration steps so that you might later modify 
    the model and perform automatic database migrations. This command will write
    alembic steps (if the database layer has been modified) to the models' alembic directory.
                """
    pause = True

    def after_run(self):
        model_file = self.model_package_dir / "models" / "records.json"

        with open(model_file) as f:
            model_data = json.load(f)

        alembic_path = self.model_dir / model_data["model"]["record-metadata"][
            "alembic"
        ].replace(".", "/")
        branch = model_data["model"]["record-metadata"]["alias"]
        self.setup_alembic(branch, alembic_path)

    def get_alembic_path(self, model_dir):
        md = model_dir
        while md != self.model_dir:
            ap = md / "alembic"
            if ap.exists():
                return ap
            md = md.parent

    def setup_alembic(self, branch, alembic_path):
        filecount = len(
            [
                x
                for x in alembic_path.iterdir()
                if x.is_file() and x.name.endswith(".py")
            ]
        )
        revision_id_prefix = branch

        def get_revision_number(stdout_str, file_suffix):
            mtch = re.search(f"(\w{{12}}){file_suffix}", stdout_str)
            if not mtch:
                raise ValueError(
                    "Revision number was not found in revision create stdout"
                )
            return mtch.group(1)

        def get_revision_names(revision_message):
            file_name = revision_message[0].lower() + revision_message[1:]
            file_name = "_" + file_name.replace(" ", "_")
            if file_name[-1] == ".":
                file_name = file_name[:-1]

            file_name = file_name[
                :30
            ]  # there seems to be maximum length for the file name
            idx = file_name.rfind("_")
            file_name = file_name[:idx]  # and all words after it are cut
            return revision_message, file_name

        def rewrite_revision_file(new_id_number, current_revision_id):
            files = list(alembic_path.iterdir())
            files_with_this_revision_id = [
                file_name
                for file_name in files
                if current_revision_id in str(file_name)
            ]

            if not files_with_this_revision_id:
                raise ValueError(
                    "Alembic file rewrite couldn't find the generated revision file"
                )

            if len(files_with_this_revision_id) > 1:
                raise ValueError(
                    "More alembic files with the same revision number found"
                )

            target_file = str(files_with_this_revision_id[0])
            new_id = f"{revision_id_prefix}_{new_id_number}"
            with open(target_file, "r") as f:
                file_text = f.read()
                file_text = file_text.replace(
                    f"revision = '{current_revision_id}'", f"revision = '{new_id}'"
                )
            with open(target_file.replace(current_revision_id, new_id), "w") as f:
                f.write(file_text)
            os.remove(target_file)

        if filecount < 2:
            # alembic has not been initialized yet ...
            self.site_support.call_invenio(
                "alembic", "upgrade", "heads", cwd=self.site_dir
            )
            # create model branch
            revision_message, file_revision_name_suffix = get_revision_names(
                f"Create {branch} branch for {self.data['model_package']}."
            )
            new_revision = get_revision_number(
                self.site_support.call_invenio(
                    "alembic",
                    "revision",
                    revision_message,
                    "-b",
                    branch,
                    "-p",
                    "dbdbc1b19cf2",
                    "--empty",
                    cwd=self.site_dir,
                    grab_stdout=True,
                ),
                file_revision_name_suffix,
            )

            rewrite_revision_file("1", new_revision)

            self.fix_sqlalchemy_utils(alembic_path)
            self.site_support.call_invenio(
                "alembic", "upgrade", "heads", cwd=self.site_dir
            )

            revision_message, file_revision_name_suffix = get_revision_names(
                "Initial revision."
            )
            new_revision = get_revision_number(
                self.site_support.call_invenio(
                    "alembic",
                    "revision",
                    revision_message,
                    "-b",
                    branch,
                    cwd=self.site_dir,
                    grab_stdout=True,
                ),
                file_revision_name_suffix,
            )

            rewrite_revision_file(
                "2", new_revision
            )  # the link to down-revision is created correctly after alembic upgrade heads on the corrected file, explicit rewrite of down-revision is not needed

            self.fix_sqlalchemy_utils(alembic_path)
            self.site_support.call_invenio(
                "alembic", "upgrade", "heads", cwd=self.site_dir
            )
        else:
            # alembic has been initialized, update heads and generate
            files = [file_path.name for file_path in alembic_path.iterdir()]

            file_numbers = []
            for file in files:
                file_number_regex = re.findall(f"(?<={revision_id_prefix}_)\d+", file)
                if file_number_regex:
                    file_numbers.append(int(file_number_regex[0]))
            new_file_number = max(file_numbers) + 1

            revision_message, file_revision_name_suffix = get_revision_names(
                "Nrp install revision."
            )
            self.site_support.call_invenio("alembic", "upgrade", "heads")
            new_revision = get_revision_number(
                self.site_support.call_invenio(
                    "alembic",
                    "revision",
                    revision_message,
                    "-b",
                    branch,
                    grab_stdout=True,
                ),
                file_revision_name_suffix,
            )
            rewrite_revision_file(new_file_number, new_revision)

            self.fix_sqlalchemy_utils(alembic_path)
            self.site_support.call_invenio("alembic", "upgrade", "heads")

    def fix_sqlalchemy_utils(self, alembic_path):
        for fn in alembic_path.iterdir():
            if not fn.name.endswith(".py"):
                continue
            data = fn.read_text()

            empty_migration = '''
def upgrade():
    """Upgrade database."""
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###'''

            if empty_migration in data:
                print(
                    f"{Fore.YELLOW}Found empty migration in file {fn}, deleting it{Style.RESET_ALL}"
                )
                fn.unlink()
                continue

            modified = False
            if "import sqlalchemy_utils" not in data:
                data = "import sqlalchemy_utils\n" + data
                modified = True
            if "import sqlalchemy_utils.types" not in data:
                data = "import sqlalchemy_utils.types\n" + data
                modified = True
            if modified:
                fn.write_text(data)

    def should_run(self):
        return True



import shutil

from oarepo_cli.site.mixins import SiteWizardStepMixin
from oarepo_cli.wizard import WizardStep


class CheckVirtualenvStep(SiteWizardStepMixin, WizardStep):
    def __init__(self, clean=False, **kwargs):
        super().__init__(**kwargs)
        self.clean = clean

    def after_run(self):
        self.site_support.check_and_create_virtualenv(clean=self.clean)
        # venv was modified, so need to get rid of compiled stuff
        if self.site_support.invenio_instance_path.exists():
            shutil.rmtree(self.site_support.invenio_instance_path)

    def should_run(self):
        print("venv ok returning", self.site_support.venv_ok())
        return not self.site_support.venv_ok()



from oarepo_cli.site.mixins import SiteWizardStepMixin
from oarepo_cli.wizard import WizardStep


class CheckDependenciesStep(SiteWizardStepMixin, WizardStep):
    def __init__(self, clean=False, require_up_to_date=False):
        super().__init__()
        self.clean = clean
        self.require_up_to_date = require_up_to_date

    def after_run(self):
        if self.require_up_to_date:
            self.site_support.require_dependencies_up_to_date()
        else:
            self.site_support.build_dependencies()

    def should_run(self):
        return (
            self.require_up_to_date
            or self.clean
            or not (self.site_support.site_dir / "requirements.txt").exists()
        )



from oarepo_cli.wizard import Wizard
from oarepo_cli.wizard.docker import DockerRunner

from ..site.add.steps.link_env import LinkEnvStep
from .steps.run_server import RunServerStep


class RunSiteWizard(Wizard):
    def __init__(self, runner: DockerRunner, *, site_support):
        self.site_support = site_support
        super().__init__(
            LinkEnvStep(),
            *runner.wrap_docker_steps(
                RunServerStep(),
            ),
        )



from oarepo_cli.model.utils import ModelWizardStep
from oarepo_cli.site.site_support import SiteSupport


class UnInstallModelStep(ModelWizardStep):
    def should_run(self):
        return True

    def after_run(self):
        sites = self.data.whole_data["sites"].keys()
        for site in sites:
            SiteSupport(self.data, site).rebuild_site(clean=True, build_ui=True)



import json
import tempfile

from oarepo_cli.site.mixins import SiteWizardStepMixin
from oarepo_cli.wizard import WizardStep


class CheckDBStep(SiteWizardStepMixin, WizardStep):
    def after_run(self):
        if self.db_status == "not_initialized":
            self.site_support.call_invenio("db", "create")
        elif self.db_status == "migration_pending":
            self.site_support.call_invenio("alembic", "upgrade", "heads")
        else:
            raise Exception(f'db error not handled: "{self.db_status}"')

    def should_run(self):
        with tempfile.NamedTemporaryFile(suffix=".json") as f:
            self.site_support.call_invenio(
                "oarepo", "check", f.name, raise_exception=True, grab_stdout=True
            )
            f.seek(0)
            data = json.load(f)
            self.db_status = data["db"]
            return self.db_status != "ok"



from oarepo_cli.model.gen.base import GeneratedFile
from oarepo_cli.model.utils import ModelWizardStep
from oarepo_cli.utils import unique_merger
from oarepo_cli.wizard import InputStep


class AddPermissionsWizardStep(ModelWizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            InputStep(
                "permissions_preset",
                prompt="""
Which permission presets do you want to use? Pre-defined values are
"'read_only' and 'everyone' but note that you may define your own presets.
Enter 'none' to not define permission preset at all (and use vanilla invenio 
permission classes).
""",
                default="read_only",
            ),
            **kwargs
        )

    def after_run(self):
        if self.data["permissions_preset"] == "none":
            return

        yaml_file: GeneratedFile = self.root.files.get("model.yaml")
        yaml = yaml_file.yaml
        unique_merger.merge(
            yaml,
            {
                "record": {
                    "permissions": {
                        "presets": [
                            x.strip()
                            for x in self.data["permissions_preset"].split(",")
                            if x.strip()
                        ]
                    }
                }
            },
        )
        yaml_file.save()

    def should_run(self):
        return True



from colorama import Fore, Style

from oarepo_cli.model.utils import ModelWizardStep
from oarepo_cli.wizard import RadioStep


class UpdateSearchIndexModelStep(ModelWizardStep):
    steps = (
        RadioStep(
            "update_opensearch",
            options={
                "run": f"{Fore.GREEN}Update opensearch index{Style.RESET_ALL}",
                "skip": f"{Fore.RED}Do not update opensearch index{Style.RESET_ALL}",
            },
            default="run",
            heading=f"""
Before the model can be used, I need to create index inside opensearch server.
This is not necessary if the model has not been changed. Should I create/update
the index? 
                            """,
            force_run=True,
        ),
    )

    def after_run(self):
        if self.data["update_opensearch"] == "run":
            self.site_support.call_invenio("oarepo", "index", "init")
            self.site_support.call_invenio("oarepo", "cf", "init")

    def should_run(self):
        return True



from oarepo_cli.site.site_support import SiteSupport
from oarepo_cli.wizard import WizardStep


class RunInvenioStep(WizardStep):
    def __init__(self, invenio_args):
        super().__init__()
        self.invenio_args = invenio_args

    def should_run(self):
        return True

    def after_run(self):
        site_support: SiteSupport = self.root.site_support
        site_support.call_invenio(*self.invenio_args, cwd=site_support.site_dir)



from oarepo_cli.model.gen.base import GeneratedFile
from oarepo_cli.model.utils import ModelWizardStep
from oarepo_cli.utils import to_python_name, unique_merger
from oarepo_cli.wizard import InputStep


class EmptyModelWizardStep(ModelWizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            InputStep(
                "model_package",
                prompt="Enter the model package",
                default=lambda data: to_python_name(data.section),
            ),
            **kwargs
        )

    def after_run(self):
        yaml_file: GeneratedFile = self.root.files.get("model.yaml")
        yaml = yaml_file.yaml
        unique_merger.merge(
            yaml,
            {
                "record": {
                    "use": ["invenio"],
                    "module": {
                        "qualified": self.data["model_package"],
                    },
                    "properties": {},
                },
                "settings": {"i18n-languages": ["en"]},
                "profiles": ["record"],
            },
        )
        yaml_file.save()

    def should_run(self):
        return True



from __future__ import annotations

from . import WizardBase
from .steps import WizardStep


class Wizard(WizardBase):
    def __init__(self, *steps: WizardStep):
        super().__init__(steps)
        self._data = None
        self.no_input = False
        self.silent = False
        self.verbose = False

    def should_run(self):
        return super().should_run()

    @property
    def data(self):
        return self._data

    def run_wizard(
        self, data, *, no_input=False, silent=False, selected_steps=None, verbose=False
    ):
        self._data = data
        self.no_input = no_input
        self.silent = silent
        self.verbose = verbose
        super().run(selected_steps=selected_steps)
        if not selected_steps:
            self.after_run()

    def after_run(self):
        pass

    def list_steps(self):
        for s in self.steps:
            print(s.name)



from pathlib import Path

from oarepo_cli.utils import ProjectWizardMixin, run_cmdline
from oarepo_cli.wizard import WizardStep


class GitHubCloneWizardStep(ProjectWizardMixin, WizardStep):
    def after_run(self):
        self.local_dir.parent.mkdir(exist_ok=True, parents=True)
        if self.data.get("github_clone_url"):
            if self.data.get("branch"):
                run_cmdline(
                    "git",
                    "clone",
                    "--branch",
                    self.data["branch"],
                    self.data["github_clone_url"],
                    str(self.local_dir),
                    cwd=self.local_dir.parent,
                )
            else:
                run_cmdline(
                    "git",
                    "clone",
                    self.data["github_clone_url"],
                    str(self.local_dir),
                    cwd=self.local_dir.parent,
                )
            # it has been checked out, so it probably is there just because we are working
            # on a library - so put it to gitignore
            self.add_to_gitignore()
        else:
            self.run_cookiecutter(
                template="https://github.com/AntoineCezar/cookiecutter-pypkg",
                config_file=f"local-{self.local_name}",
                output_dir=self.data.project_dir / "local",
                extra_context={
                    "project_name": "Python Project",
                    "project_slug": self.local_name,
                    "package_name": "{{ cookiecutter.project_slug.replace('-', '_') }}",
                    "test_runner": "pytest",
                    "build_docs": "n",
                    "build_rpm": "n",
                    "exemple": "n",
                    "git_init": "n",
                },
            )
            # this is a newly created, probably will stay here - so do not put it to gitignore

    def should_run(self):
        return not self.local_dir.exists()

    @property
    def local_name(self):
        return self.data.section_path[-1]

    @property
    def local_dir(self):
        return self.data.project_dir / self.data["local_dir"]

    def add_to_gitignore(self):
        gitignore: Path = self.local_dir.parent / ".gitignore"
        if gitignore.exists():
            content = gitignore.read_text()
        else:
            content = ""
        content = [x.strip() for x in content.split("\n")]
        if self.local_name not in content:
            content.append(self.local_name)
            gitignore.write_text("\n".join(content))



import time

import click
import psutil


@click.command(name="kill", hidden=True)
@click.argument("pid", type=int)
def kill_command(pid):
    kill(pid)


def kill(pid):
    try:
        process = psutil.Process(pid)
    except Exception as e:
        print(e)
        return

    to_kill = [process] + list(process.children(True))
    to_kill.reverse()
    print(f"Going to kill {[x.pid for x in to_kill]}")

    # first round - send terminate
    for c in to_kill:
        if c.is_running():
            try:
                print(f"Terminating {c.pid} {' '.join(c.cmdline())}")
            except:
                pass
            c.terminate()

    # if all killed, return
    for c in to_kill:
        if c.is_running():
            break
    else:
        return  # all terminated

    # wait a bit and try again
    time.sleep(2)

    # if all killed after waiting, return
    for c in to_kill:
        if c.is_running():
            break
    else:
        return  # all terminated

    # send the rest sigkill
    for c in to_kill:
        if c.is_running():
            try:
                print(f"Killing {c.pid} {c.cmdline()}")
            except:
                pass
            c.kill()



from oarepo_cli.wizard import Wizard
from oarepo_cli.wizard.docker import DockerRunner

from ..site.site_support import SiteSupport
from .steps.format_javascript import FormatJavascriptStep
from .steps.format_jinja import FormatJinjaStep
from .steps.format_python import FormatPythonStep


class FormatWizard(Wizard):
    def __init__(self, runner: DockerRunner, site_support: SiteSupport):
        self.site_support = site_support
        steps = []
        steps.extend(
            runner.wrap_docker_steps(
                FormatPythonStep(),
                FormatJinjaStep(),
                FormatJavascriptStep(),
            )
        )
        super().__init__(*steps)



from oarepo_cli.site.site_support import SiteSupport
from oarepo_cli.utils import ProjectWizardMixin, SiteMixin
from oarepo_cli.wizard import WizardStep


class BuildAssetsUIStep(SiteMixin, ProjectWizardMixin, WizardStep):
    def should_run(self):
        return True

    def after_run(self):
        sites = self.data["sites"]
        for site_name in sites:
            support = SiteSupport(self.data, site_name)
            support.build_ui()



from oarepo_cli.model.gen.base import GeneratedFile
from oarepo_cli.model.utils import ModelWizardStep
from oarepo_cli.package_versions import (
    OAREPO_MODEL_BUILDER_DRAFTS_FILES_VERSION,
    OAREPO_MODEL_BUILDER_DRAFTS_VERSION,
)
from oarepo_cli.utils import unique_merger
from oarepo_cli.wizard import RadioStep


class AddDraftsWizardStep(ModelWizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            RadioStep(
                "use_drafts",
                heading="Should records be first uploaded as editable drafts before they are published?",
                options={
                    "yes": "yes",
                    "no": "no",
                },
                default="yes",
            ),
            **kwargs,
        )

    def after_run(self):
        if self.data["use_drafts"] != "yes":
            return

        yaml_file: GeneratedFile = self.root.files.get("model.yaml")
        yaml = yaml_file.yaml
        unique_merger.merge(
            yaml,
            {
                "record": {"draft": {}},
                "profiles": ["draft"],
                "plugins": {
                    "packages": [
                        f"oarepo-model-builder-drafts{OAREPO_MODEL_BUILDER_DRAFTS_VERSION}"
                    ]
                },
            },
        )

        if self.data["use_files"] == "yes":
            unique_merger.merge(
                yaml,
                {
                    "record": {"draft-files": {}},
                    "profiles": ["draft_files"],
                    "plugins": {
                        "packages": [
                            f"oarepo-model-builder-drafts-files{OAREPO_MODEL_BUILDER_DRAFTS_FILES_VERSION}"
                        ]
                    },
                },
            )

        yaml_file.save()

    def should_run(self):
        return True



from oarepo_cli.site.site_support import SiteSupport
from oarepo_cli.wizard import WizardStep


class RunOARepoStep(WizardStep):
    def __init__(self, oarepo_args):
        super().__init__()
        self.oarepo_args = oarepo_args

    def should_run(self):
        return True

    def after_run(self):
        site_support: SiteSupport = self.root.site_support
        site_support.call_invenio(
            "oarepo", *self.oarepo_args, cwd=site_support.site_dir
        )



import threading

from oarepo_cli.site.site_support import SiteSupport
from oarepo_cli.utils import exec_nrp_in_docker, run_nrp_in_docker_compose


class DockerDevelopmentRunner:
    def __init__(self, site_support: SiteSupport):
        self.site_support: SiteSupport = site_support

    def start(self):
        self.development_server_thread = threading.Thread(
            target=lambda: self.run_docker_develop(), daemon=True
        )
        self.development_server_thread.start()

    @property
    def develop_container_name(self):
        return self.site_support.site_name + "-repo-develop"

    def run_docker_develop(self):
        run_nrp_in_docker_compose(
            self.site_support.site_dir,
            "develop",
            "--site",
            self.site_support.site_name,
            "--step",
            "DevelopStep",
            interactive=False,
            no_input=True,
            name=self.develop_container_name,
        )

    def stop(self):
        self.send_command("stop")

    def restart_python(self):
        self.send_command("server")

    def restart_ui(self):
        self.send_command("ui")

    def send_command(self, command):
        exec_nrp_in_docker(
            self.site_support.site_dir,
            self.develop_container_name,
            "develop",
            "--command",
            command,
            "--site",
            self.site_support.site_name,
            interactive=False,
        )



import click as click

from oarepo_cli.utils import with_config

from ..site.site_support import SiteSupport
from ..wizard.docker import DockerRunner
from .wizard import InvenioWizard


@click.command(
    name="invenio",
    help="""
Run the invenio command, possibly inside a docker""",
    context_settings=dict(
        ignore_unknown_options=True,
    ),
)
@click.option("--site", required=False)
@click.argument("invenio_args", nargs=-1, type=click.UNPROCESSED)
@with_config()
def invenio_command(
    cfg=None,
    no_input=False,
    silent=False,
    step=None,
    steps=False,
    verbose=False,
    site=None,
    invenio_args=None,
    **kwargs,
):
    site_support = SiteSupport(cfg, site)
    runner = DockerRunner(cfg, no_input)
    wizard = InvenioWizard(runner, site_support=site_support, invenio_args=invenio_args)

    if steps:
        wizard.list_steps()
        return

    wizard.run_wizard(
        cfg, no_input=no_input, silent=silent, selected_steps=step, verbose=verbose
    )



from oarepo_cli.site.site_support import SiteSupport
from oarepo_cli.wizard import WizardStep


class UpgradeSiteStep(WizardStep):
    def __init__(self, site=None):
        super().__init__()
        self.site = site

    def should_run(self):
        return True

    def after_run(self):
        if not hasattr(self.root, "site_support"):
            site_support = SiteSupport(self.data, self.site)
        else:
            site_support = self.root.site_support
        site_support.rebuild_site(clean=True, build_ui=True)



from {{cookiecutter.app_package}}.proxies import current_ui


def create_blueprint(app):
    """Blueprint for the routes and resources provided by current ui's resource."""
    with app.app_context():
        app.extensions['oarepo_ui'].register_resource(current_ui.resource)
        return current_ui.resource.as_blueprint()



from __future__ import annotations

from .base import WizardBase
from .steps import InputStep, RadioStep, StaticStep, WizardStep
from .wizard import Wizard

__all__ = ["InputStep", "StaticStep", "RadioStep", "WizardStep", "WizardBase", "Wizard"]



import click

from oarepo_cli.build_command.cli import build_command
from oarepo_cli.develop.cli import develop_command
from oarepo_cli.format.cli import format_sources_command
from oarepo_cli.initialize.cli import initialize_command
from oarepo_cli.invenio.cli import invenio_command
from oarepo_cli.kill import kill_command
from oarepo_cli.local import local_commands
from oarepo_cli.model import model_commands
from oarepo_cli.oarepo.cli import oarepo_command
from oarepo_cli.run.cli import run_server_command
from oarepo_cli.site import site_commands
from oarepo_cli.ui import ui_commands
from oarepo_cli.upgrade.cli import upgrade_command
from oarepo_cli.watch.cli import watch_command


@click.group()
def run(*args, **kwargs):
    pass


run.add_command(initialize_command)
run.add_command(site_commands)
run.add_command(model_commands)
run.add_command(local_commands)
run.add_command(ui_commands)
run.add_command(run_server_command)
run.add_command(upgrade_command)
run.add_command(develop_command)
run.add_command(watch_command)
run.add_command(format_sources_command)
run.add_command(kill_command)
run.add_command(build_command)
run.add_command(invenio_command)
run.add_command(oarepo_command)

if __name__ == "__main__":
    run()



import click as click

from oarepo_cli.model.model_support import ModelSupport
from oarepo_cli.utils import commit_git, with_config
from oarepo_cli.wizard.docker import DockerRunner

from .wizard import CompileModelWizard


@click.command(
    name="compile",
    help="""
Compile model yaml file to invenio sources. Required arguments:
    <name>   ... name of the already existing model
""",
)
@click.argument("name", required=True)
@with_config(config_section=lambda name, **kwargs: ["models", name])
def compile_model(
    cfg=None,
    no_input=False,
    silent=False,
    step=None,
    steps=False,
    verbose=False,
    **kwargs,
):
    commit_git(
        cfg.project_dir,
        f"before-model-compile-{cfg.section}",
        f"Committed automatically before model {cfg.section} has been compiled",
    )
    runner = DockerRunner(cfg, no_input)
    wizard = CompileModelWizard(runner)
    wizard.model_support = ModelSupport(cfg)
    if steps:
        wizard.list_steps()
        return

    wizard.run_wizard(
        cfg, no_input=no_input, silent=silent, selected_steps=step, verbose=verbose
    )
    commit_git(
        cfg.project_dir,
        f"after-model-compile-{cfg.section}",
        f"Committed automatically after model {cfg.section} has been compiled",
    )



from oarepo_cli.oarepo.steps.run_oarepo import RunOARepoStep
from oarepo_cli.wizard import Wizard
from oarepo_cli.wizard.docker import DockerRunner


class OARepoWizard(Wizard):
    def __init__(self, runner: DockerRunner, *, site_support, oarepo_args):
        self.site_support = site_support

        super().__init__(
            *runner.wrap_docker_steps(
                RunOARepoStep(oarepo_args),
            ),
        )



import functools
from pathlib import Path
from typing import Generic, TypeVar

import importlib_metadata

T = TypeVar("T")


class FileFormat(Generic[T]):
    def load(self, text: str) -> T:
        raise NotImplementedError("Implement this method")

    def dump(self, data: T) -> str:
        raise NotImplementedError("Implement this method")


class GeneratedFile:
    def __init__(self, files, fname: Path):
        self._files = files
        self._fname = fname
        if self._fname.exists():
            self._content = self._fname.read_text()
        else:
            self._content = ""
        self._active_format = "text"
        self._formats = {}

    def __getattr__(self, format_name):
        self._content = self._convert_to(format_name)
        self._active_format = format_name
        return self._content

    def _convert_to(self, format_name):
        if self._active_format == format_name:
            return self._content
        current_format = self._files.get_format(self._active_format)
        new_format = self._files.get_format(format_name)
        data = current_format.dump(self._content)
        return new_format.load(data)

    def save(self):
        self._fname.parent.mkdir(parents=True, exist_ok=True)
        self._fname.write_text(self._convert_to("text"))


class GeneratedFiles:
    def __init__(self, base_path: Path):
        self._base_path = base_path
        self._files = {}

    @functools.lru_cache
    def get_format(self, format_name) -> FileFormat:
        return list(
            importlib_metadata.entry_points().select(
                group="nrp_cli.file_formats", name=format_name
            )
        )[0].load()()

    def get(self, fname):
        if fname not in self._files:
            self._files[fname] = GeneratedFile(self, self._base_path / fname)
        return self._files[fname]



from flask import current_app
from werkzeug.local import LocalProxy

current_ui = LocalProxy(lambda: current_app.extensions["{{cookiecutter.app_package}}"])
"""Proxy to the instantiated ui extension."""



from __future__ import annotations

import datetime
import os
import re

from oarepo_cli.site.mixins import SiteWizardStepMixin
from oarepo_cli.templates import get_cookiecutter_template
from oarepo_cli.utils import ProjectWizardMixin, commit_git
from oarepo_cli.wizard import InputStep, RadioStep, StaticStep, WizardStep


class InstallSiteStep(SiteWizardStepMixin, ProjectWizardMixin, WizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            InputStep(
                "repository_name",
                prompt="""Enter the repository name ("title" of the HTML site)""",
                default=lambda data: re.sub("[_-]", " ", self.site_dir.name).title(),
            ),
            InputStep(
                "www",
                prompt="""Enter the WWW address on which the repository will reside""",
            ),
            InputStep(
                "author_name",
                prompt="""Author name""",
                default=os.environ.get("USERNAME") or os.environ.get("USER"),
            ),
            InputStep("author_email", prompt="""Author email"""),
            InputStep(
                "year",
                prompt="""Year (for copyright)""",
                default=datetime.datetime.today().strftime("%Y"),
            ),
            InputStep("copyright_holder", prompt="""Copyright holder"""),
            RadioStep(
                "use_oarepo_vocabularies",
                options={"yes": "Yes", "no": "No"},
                default="yes",
                heading=f"""
                    Are you planning to use extended vocabularies (extra fields on vocabularies, hierarchy in vocabulary items)? If in doubt, select 'yes'.
                        """,
            ),
            StaticStep(
                heading="""I have all the information to generate the site.
        To do so, I'll call the cookiecutter to install the size template. If anything goes wrong, please fix the problem
        and run the wizard again.
                    """,
            ),
            heading="""
Now I will add site sources, that can be used to change the overall CSS style and
configure your repository. Please fill in the required values. 
If not sure, keep the default values.""",
            **kwargs,
        )

    def on_before_run(self):
        self.data.setdefault(
            "transifex_project", self.data.get("config.project_package", "")
        )
        return True

    def after_run(self):
        site_name = self.site_dir.name
        site_dir = self.site_dir
        if not site_dir.parent.exists():
            site_dir.parent.mkdir(parents=True)

        self.run_cookiecutter(
            template=get_cookiecutter_template("site"),
            config_file=f"site-{site_name}",
            output_dir=str(site_dir.parent),
            extra_context=dict(
                project_name=self.data["repository_name"],
                project_shortname=self.site_dir.name,
                project_site=self.data["www"],
                jsonschemas_host=get_host_name(self.data["www"]),
                github_repo="",
                description=f"{self.data['repository_name']} OARepo Instance",
                author_name=self.data["author_name"],
                author_email=self.data["author_email"],
                year=self.data["year"],
                python_version="3.9",
                database="postgresql",
                search="opensearch2",
                file_storage="S3",
                development_tools="yes",
                site_code="yes",
                use_oarepo_vocabularies=self.data["use_oarepo_vocabularies"],
            ),
        )

        commit_git(
            self.data.project_dir,
            f"after-site-cookiecutter-{self.data.section}",
            f"Committed automatically after site {self.data.section} cookiecutter has been called",
        )

    def should_run(self):
        return not (self.site_dir / "variables").exists()


def get_host_name(url):
    if url.startswith("http://"):
        url = url[7:]
    if url.startswith("https://"):
        url = url[8:]
    return url.split("/", maxsplit=1)[0]



import os
import shutil
from pathlib import Path

from oarepo_cli.templates import get_cookiecutter_template
from oarepo_cli.utils import ProjectWizardMixin, commit_git
from oarepo_cli.wizard import WizardStep


def keep_existing_copy(src, dst, *, follow_symlinks=True):
    if os.path.isdir(dst):
        _dst = os.path.join(dst, os.path.basename(src))
    else:
        _dst = dst
    if Path(_dst).exists():
        return _dst
    return shutil.copy2(src, dst, follow_symlinks=follow_symlinks)


class CreateMonorepoStep(ProjectWizardMixin, WizardStep):
    def after_run(self):
        project_dir, repo_name, repo_out = self._repo_params
        self.run_cookiecutter(
            template=get_cookiecutter_template("repo"),
            config_file="monorepo",
            output_dir=str(repo_out),
            extra_context={
                **self.data,
                "repo_name": repo_name,
                "repo_human_name": repo_name,
            },
        )
        for f in (repo_out / repo_name).iterdir():
            shutil.move(f, project_dir, copy_function=keep_existing_copy)
        os.rmdir(repo_out / repo_name)
        os.rmdir(repo_out)
        commit_git(
            project_dir,
            "after-monorepo",
            "Committed automatically after monorepo has been created",
        )
        return True

    @property
    def _repo_params(self):
        project_dir = self.data.project_dir
        repo_name = project_dir.name
        repo_out = project_dir.parent / (project_dir.name + "-tmp")
        return project_dir, repo_name, repo_out

    def should_run(self):
        project_dir, repo_name, repo_out = self._repo_params
        return not (project_dir / "nrp").exists()
